import random
import re
import mysql.connector
import Levenshtein

class ModellBeta:
    def __init__(self, db_config):
        # Csatlakozás a MySQL adatbázishoz
        try:
            self.db_connection = mysql.connector.connect(**db_config)
            self.cursor = self.db_connection.cursor()
        except mysql.connector.Error as err:
            print(f"Error: {err}")
            raise

        # Eltároljuk a válaszokat és kulcsszavakat az adatbázisból
        self.responses = self.load_responses()
        self.topics = self.load_topics()

        # Emlékezet és alapértelmezett válaszok
        self.default_responses = [
            "Sajnos nem értem, fogalmazd át kérlek!", 
            "Eltudnád mondani másképp, hátha megértem!", 
            "Sajnos nem tudok erre válaszolni.", 
            "Megpróbálnád máshogy leírni?"
        ]
        self.memory = []  # Emlékezik a beszélgetésekre

    def load_responses(self):
        """Betölti az adatbázisból a válaszokat"""
        self.cursor.execute("SELECT id, keyword, response, score FROM responses")
        result = self.cursor.fetchall()
        
        responses = {}
        for id, keyword, response, score in result:
            responses[keyword] = {
                "response": response,
                "score": score
            }
        return responses

    def tokenize(self, message):
        """Tokenizálás a szöveg feldarabolása szavakra"""
        tokens = re.findall(r'\w+', message.lower())  # Minden szót különválasztunk
        return tokens

    def match_responses(self, tokens):
        """Megkeresi a válaszokat az adatbázisban"""
        best_match = None
        best_score = float('inf')
        
        for keyword, data in self.responses.items():
            keyword_tokens = keyword.split()  # Kulcsszavak szétbontása
            distance = sum(Levenshtein.distance(token, keyword_token) for token in tokens for keyword_token in keyword_tokens)
            
            if distance < best_score:
                best_score = distance
                best_match = keyword
        
        if best_score < 5:  # Ha a távolság elég kicsi
            return best_match
        return None

    def match_topics(self, tokens):
        """Megkeresi a témákat az adatbázisban"""
        best_topic_match = None
        
        for topic in self.topics.values():
            for keyword in topic["keywords"]:
                if all(word in tokens for word in keyword.lower().split()):
                    best_topic_match = topic["response"]
                    break
            if best_topic_match:
                break
        return best_topic_match

    def search_in_history(self, message):
        """Megkeresi a beszélgetés előzményeit és a score értékeket"""
        self.cursor.execute("""
            SELECT user_message, bot_response, score
            FROM conversation_history
            WHERE score IS NOT NULL
            ORDER BY score DESC
            LIMIT 100
        """)
        rows = self.cursor.fetchall()

        best_match = None
        best_score = 0
        for user_msg, bot_resp, score in rows:
            if score < 4:
                continue
            similarity = self.calculate_similarity(message, user_msg)
            combined_score = similarity * score

            if combined_score > best_score:
                best_score = combined_score
                best_match = bot_resp

        return best_match

    def calculate_similarity(self, msg1, msg2):
        """Hasonlóság kalkuláció tokenek alapján"""
        tokens1 = set(self.tokenize(msg1))
        tokens2 = set(self.tokenize(msg2))
        if not tokens1 or not tokens2:
            return 0
        common = tokens1.intersection(tokens2)
        return len(common) / max(len(tokens1), len(tokens2))

    def is_bad_response(self, user_message, candidate_response):
        """Ellenőrzi, hogy korábban ezt a választ már megadta-e ugyanilyen kérdésre és arra 0 pont járt-e"""
        self.cursor.execute("""
            SELECT score FROM conversation_history
            WHERE user_message = %s AND bot_response = %s
            ORDER BY id DESC LIMIT 1
        """, (user_message, candidate_response))

        result = self.cursor.fetchone()
        if result and result[0] == 0:
            return True
        return False

    def get_response(self, message):
        tokens = self.tokenize(message)
    
        # 1. Responses adatbázis alapján
        matched_keyword = self.match_responses(tokens)
        if matched_keyword:
            candidate = self.responses[matched_keyword]["response"]
            if not self.is_bad_response(message, candidate):
                bot_response = candidate
            else:
                bot_response = None
        else:
            bot_response = None
    
        # 2. Témák alapján
        if not bot_response:
            topic_response = self.match_topics(tokens)
            if topic_response and not self.is_bad_response(message, topic_response):
                bot_response = topic_response
    
        # 3. Előzményekben keresés pozitív score alapján
        if not bot_response:
            history_response = self.search_in_history(message)
            if history_response and not self.is_bad_response(message, history_response):
                bot_response = history_response
    
        # 4. Ha nincs jó válasz vagy mindet kizártuk a pontozás miatt
        if not bot_response:
            bot_response = random.choice(self.default_responses)
    
        # Mentés adatbázisba
        self.store_conversation(message, bot_response)
        return bot_response

    def store_conversation(self, user_message, bot_response):
        """Beszélgetések tárolása az adatbázisba"""
        sql = "INSERT INTO conversation_history (user_message, bot_response) VALUES (%s, %s)"
        self.cursor.execute(sql, (user_message, bot_response))
        self.db_connection.commit()

    def load_topics(self):
        """Témák betöltése"""
        self.cursor.execute("""
            SELECT t.id, t.name, t.response, k.keyword 
            FROM topics t 
            JOIN topic_keywords k ON t.id = k.topic_id
        """)
        rows = self.cursor.fetchall()
        topics = {}
        for topic_id, name, response, keyword in rows:
            topics.setdefault(topic_id, {"name": name, "response": response, "keywords": []})
            topics[topic_id]["keywords"].append(keyword)
        return topics


